@@ -12,10 +12,13 @@
 #include <algorithm>
 #include <memory>
 #include <utility>
+#include <limits>
 #include <vector>
 
 #include "categorical.h"
 #include "column_matrix.h"
+#include "../tree/hist/expand_entry.h"
+#include "xgboost/generic_parameters.h"
 #include "xgboost/tree_model.h"
 
 namespace xgboost {
@@ -50,23 +53,23 @@ class PartitionBuilder {
   // Handle dense columns
   // Analog of std::stable_partition, but in no-inplace manner
   template <bool default_left, bool any_missing, typename ColumnType, typename Predicate>
-  inline std::pair<size_t, size_t> PartitionKernel(const ColumnType& column,
+  inline std::pair<size_t, size_t> PartitionKernel(ColumnType* p_column,
                                                    common::Span<const size_t> row_indices,
                                                    common::Span<size_t> left_part,
                                                    common::Span<size_t> right_part,
                                                    size_t base_rowid, Predicate&& pred) {
+    auto& column = *p_column;
     size_t* p_left_part = left_part.data();
     size_t* p_right_part = right_part.data();
     size_t nleft_elems = 0;
     size_t nright_elems = 0;
-    auto state = column.GetInitialState(row_indices.front() - base_rowid);
 
     auto p_row_indices = row_indices.data();
     auto n_samples = row_indices.size();
 
     for (size_t i = 0; i < n_samples; ++i) {
       auto rid = p_row_indices[i];
-      const int32_t bin_id = column.GetBinIdx(rid - base_rowid, &state);
+      const int32_t bin_id = column[rid - base_rowid];
       if (any_missing && bin_id == ColumnType::kMissingId) {
         if (default_left) {
           p_left_part[nleft_elems++] = rid;
@@ -105,16 +108,17 @@ class PartitionBuilder {
   }
 
   template <typename BinIdxType, bool any_missing, bool any_cat>
-  void Partition(const size_t node_in_set, const size_t nid, const common::Range1d range,
-                 const int32_t split_cond, GHistIndexMatrix const& gmat,
-                 const ColumnMatrix& column_matrix, const RegTree& tree, const size_t* rid) {
+  void Partition(const size_t node_in_set, std::vector<xgboost::tree::CPUExpandEntry> const &nodes,
+                 const common::Range1d range,
+                 const bst_bin_t split_cond, GHistIndexMatrix const& gmat,
+                 const common::ColumnMatrix& column_matrix,
+                 const RegTree& tree, const size_t* rid) {
     common::Span<const size_t> rid_span(rid + range.begin(), rid + range.end());
     common::Span<size_t> left = GetLeftBuffer(node_in_set, range.begin(), range.end());
     common::Span<size_t> right = GetRightBuffer(node_in_set, range.begin(), range.end());
-    const bst_uint fid = tree[nid].SplitIndex();
-    const bool default_left = tree[nid].DefaultLeft();
-    const auto column_ptr = column_matrix.GetColumn<BinIdxType, any_missing>(fid);
-
+    std::size_t nid = nodes[node_in_set].nid;
+    bst_feature_t fid = tree[nid].SplitIndex();
+    bool default_left = tree[nid].DefaultLeft();
     bool is_cat = tree.GetSplitTypes()[nid] == FeatureType::kCategorical;
     auto node_cats = tree.NodeCats(nid);
 
@@ -122,20 +126,25 @@ class PartitionBuilder {
     auto const& cut_values = gmat.cut.Values();
     auto const& cut_ptrs = gmat.cut.Ptrs();
 
-    auto pred = [&](auto ridx, auto bin_id) {
+    auto gidx_calc = [&](auto ridx) {
+      auto begin = gmat.RowIdx(ridx);
+      if (gmat.IsDense()) {
+        return static_cast<bst_bin_t>(index[begin + fid]);
+      }
+      auto end = gmat.RowIdx(ridx + 1);
+      auto f_begin = cut_ptrs[fid];
+      auto f_end = cut_ptrs[fid + 1];
+      // bypassing the column matrix as we need the cut value instead of bin idx for categorical
+      // features.
+      return BinarySearchBin(begin, end, index, f_begin, f_end);
+    };
+
+    auto pred_hist = [&](auto ridx, auto bin_id) {
       if (any_cat && is_cat) {
-        auto begin = gmat.RowIdx(ridx);
-        auto end = gmat.RowIdx(ridx + 1);
-        auto f_begin = cut_ptrs[fid];
-        auto f_end = cut_ptrs[fid + 1];
-        // bypassing the column matrix as we need the cut value instead of bin idx for categorical
-        // features.
-        auto gidx = BinarySearchBin(begin, end, index, f_begin, f_end);
-        bool go_left;
-        if (gidx == -1) {
-          go_left = default_left;
-        } else {
-          go_left = Decision(node_cats, cut_values[gidx], default_left);
+        auto gidx = gidx_calc(ridx);
+        bool go_left = default_left;
+        if (gidx > -1) {
+          go_left = Decision(node_cats, cut_values[gidx]);
         }
         return go_left;
       } else {
@@ -143,67 +152,51 @@ class PartitionBuilder {
       }
     };
 
-    std::pair<size_t, size_t> child_nodes_sizes;
-    if (column_ptr->GetType() == xgboost::common::kDenseColumn) {
-      const common::DenseColumn<BinIdxType, any_missing>& column =
-            static_cast<const common::DenseColumn<BinIdxType, any_missing>& >(*(column_ptr.get()));
-      if (default_left) {
-        child_nodes_sizes = PartitionKernel<true, any_missing>(column, rid_span, left, right,
-                                                               gmat.base_rowid, pred);
-      } else {
-        child_nodes_sizes = PartitionKernel<false, any_missing>(column, rid_span, left, right,
-                                                                gmat.base_rowid, pred);
+    auto pred_approx = [&](auto ridx) {
+      auto gidx = gidx_calc(ridx);
+      bool go_left = default_left;
+      if (gidx > -1) {
+        if (is_cat) {
+          go_left = Decision(node_cats, cut_values[gidx]);
+        } else {
+          go_left = cut_values[gidx] <= nodes[node_in_set].split.split_value;
+        }
       }
+      return go_left;
+    };
+
+    std::pair<size_t, size_t> child_nodes_sizes;
+    if (!column_matrix.IsInitialized()) {
+      child_nodes_sizes = PartitionRangeKernel(rid_span, left, right, pred_approx);
     } else {
-      CHECK_EQ(any_missing, true);
-      const common::SparseColumn<BinIdxType>& column
-        = static_cast<const common::SparseColumn<BinIdxType>& >(*(column_ptr.get()));
-      if (default_left) {
-        child_nodes_sizes = PartitionKernel<true, any_missing>(column, rid_span, left, right,
-                                                               gmat.base_rowid, pred);
+      if (column_matrix.GetColumnType(fid) == xgboost::common::kDenseColumn) {
+        auto column = column_matrix.DenseColumn<BinIdxType, any_missing>(fid);
+        if (default_left) {
+          child_nodes_sizes = PartitionKernel<true, any_missing>(&column, rid_span, left, right,
+                                                                 gmat.base_rowid, pred_hist);
+        } else {
+          child_nodes_sizes = PartitionKernel<false, any_missing>(&column, rid_span, left, right,
+                                                                  gmat.base_rowid, pred_hist);
+        }
       } else {
-        child_nodes_sizes = PartitionKernel<false, any_missing>(column, rid_span, left, right,
-                                                                gmat.base_rowid, pred);
+        CHECK_EQ(any_missing, true);
+        auto column =
+            column_matrix.SparseColumn<BinIdxType>(fid, rid_span.front() - gmat.base_rowid);
+        if (default_left) {
+          child_nodes_sizes = PartitionKernel<true, any_missing>(&column, rid_span, left, right,
+                                                                 gmat.base_rowid, pred_hist);
+        } else {
+          child_nodes_sizes = PartitionKernel<false, any_missing>(&column, rid_span, left, right,
+                                                                  gmat.base_rowid, pred_hist);
+        }
       }
     }
 
     const size_t n_left  = child_nodes_sizes.first;
     const size_t n_right = child_nodes_sizes.second;
 
-    SetNLeftElems(node_in_set, range.begin(), range.end(), n_left);
-    SetNRightElems(node_in_set, range.begin(), range.end(), n_right);
-  }
-
-  /**
-   * \brief Partition tree nodes with specific range of row indices.
-   *
-   * \tparam Pred       Predicate for whether a row should be partitioned to the left node.
-   *
-   * \param node_in_set The index of node in current batch of nodes.
-   * \param nid         The cannonical node index (node index in the tree).
-   * \param range       The range of input row index.
-   * \param fidx        Feature index.
-   * \param p_row_set_collection Pointer to rows that are  being partitioned.
-   * \param pred        A callback function that returns whether current row should be
-   *                    partitioned to the left node, it should accept the row index as
-   *                    input and returns a boolean value.
-   */
-  template <typename Pred>
-  void PartitionRange(const size_t node_in_set, const size_t nid, common::Range1d range,
-                      bst_feature_t fidx, common::RowSetCollection* p_row_set_collection,
-                      Pred pred) {
-    auto& row_set_collection = *p_row_set_collection;
-    const size_t* p_ridx = row_set_collection[nid].begin;
-    common::Span<const size_t> ridx(p_ridx + range.begin(), p_ridx + range.end());
-    common::Span<size_t> left = this->GetLeftBuffer(node_in_set, range.begin(), range.end());
-    common::Span<size_t> right = this->GetRightBuffer(node_in_set, range.begin(), range.end());
-    std::pair<size_t, size_t> child_nodes_sizes = PartitionRangeKernel(ridx, left, right, pred);
-
-    const size_t n_left = child_nodes_sizes.first;
-    const size_t n_right = child_nodes_sizes.second;
-
-    this->SetNLeftElems(node_in_set, range.begin(), range.end(), n_left);
-    this->SetNRightElems(node_in_set, range.begin(), range.end(), n_right);
+    SetNLeftElems(node_in_set, range.begin(), n_left);
+    SetNRightElems(node_in_set, range.begin(), n_right);
   }
 
   // allocate thread local memory, should be called for each specific task
@@ -225,12 +218,12 @@ class PartitionBuilder {
     return { mem_blocks_.at(task_idx)->Right(), end - begin };
   }
 
-  void SetNLeftElems(int nid, size_t begin, size_t end, size_t n_left) {
+  void SetNLeftElems(int nid, size_t begin, size_t n_left) {
     size_t task_idx = GetTaskIdx(nid, begin);
     mem_blocks_.at(task_idx)->n_left = n_left;
   }
 
-  void SetNRightElems(int nid, size_t begin, size_t end, size_t n_right) {
+  void SetNRightElems(int nid, size_t begin, size_t n_right) {
     size_t task_idx = GetTaskIdx(nid, begin);
     mem_blocks_.at(task_idx)->n_right = n_right;
   }
@@ -254,7 +247,7 @@ class PartitionBuilder {
         n_left += mem_blocks_[j]->n_left;
       }
       size_t n_right = 0;
-      for (size_t j = blocks_offsets_[i]; j < blocks_offsets_[i+1]; ++j) {
+      for (size_t j = blocks_offsets_[i]; j < blocks_offsets_[i + 1]; ++j) {
         mem_blocks_[j]->n_offset_right = n_left + n_right;
         n_right += mem_blocks_[j]->n_right;
       }
@@ -279,6 +272,30 @@ class PartitionBuilder {
     return blocks_offsets_[nid] + begin / BlockSize;
   }
 
+  // Copy row partitions into global cache for reuse in objective
+  template <typename Sampledp>
+  void LeafPartition(Context const* ctx, RegTree const& tree, RowSetCollection const& row_set,
+                     std::vector<bst_node_t>* p_position, Sampledp sampledp) const {
+    auto& h_pos = *p_position;
+    h_pos.resize(row_set.Data()->size(), std::numeric_limits<bst_node_t>::max());
+
+    auto p_begin = row_set.Data()->data();
+    ParallelFor(row_set.Size(), ctx->Threads(), [&](size_t i) {
+      auto const& node = row_set[i];
+      if (node.node_id < 0) {
+        return;
+      }
+      CHECK(tree[node.node_id].IsLeaf());
+      if (node.begin) {  // guard for empty node.
+        size_t ptr_offset = node.end - p_begin;
+        CHECK_LE(ptr_offset, row_set.Data()->size()) << node.node_id;
+        for (auto idx = node.begin; idx != node.end; ++idx) {
+          h_pos[*idx] = sampledp(*idx) ? ~node.node_id : node.node_id;
+        }
+      }
+    });
+  }
+
  protected:
   struct BlockInfo{
     size_t n_left;
