@@ -2,14 +2,17 @@
 
 require(xgboost)
 require(data.table)
+require(titanic)
 
 context("callbacks")
 
-data(agaricus.train, package='xgboost')
-data(agaricus.test, package='xgboost')
+data(agaricus.train, package = 'xgboost')
+data(agaricus.test, package = 'xgboost')
 train <- agaricus.train
 test <- agaricus.test
 
+n_threads <- 2
+
 # add some label noise for early stopping tests
 add.noise <- function(label, frac) {
   inoise <- sample(length(label), length(label) * frac)
@@ -19,26 +22,27 @@ add.noise <- function(label, frac) {
 set.seed(11)
 ltrain <- add.noise(train$label, 0.2)
 ltest <- add.noise(test$label, 0.2)
-dtrain <- xgb.DMatrix(train$data, label = ltrain)
-dtest <- xgb.DMatrix(test$data, label = ltest)
-watchlist = list(train=dtrain, test=dtest)
+dtrain <- xgb.DMatrix(train$data, label = ltrain, nthread = n_threads)
+dtest <- xgb.DMatrix(test$data, label = ltest, nthread = n_threads)
+watchlist <- list(train = dtrain, test = dtest)
 
 
-err <- function(label, pr) sum((pr > 0.5) != label)/length(label)
+err <- function(label, pr) sum((pr > 0.5) != label) / length(label)
 
-param <- list(objective = "binary:logistic", max_depth = 2, nthread = 2)
+param <- list(objective = "binary:logistic", eval_metric = "error",
+              max_depth = 2, nthread = n_threads)
 
 
 test_that("cb.print.evaluation works as expected", {
 
-  bst_evaluation <- c('train-auc'=0.9, 'test-auc'=0.8)
+  bst_evaluation <- c('train-auc' = 0.9, 'test-auc' = 0.8)
   bst_evaluation_err <- NULL
   begin_iteration <- 1
   end_iteration <- 7
 
-  f0 <- cb.print.evaluation(period=0)
-  f1 <- cb.print.evaluation(period=1)
-  f5 <- cb.print.evaluation(period=5)
+  f0 <- cb.print.evaluation(period = 0)
+  f1 <- cb.print.evaluation(period = 1)
+  f5 <- cb.print.evaluation(period = 5)
 
   expect_false(is.null(attr(f1, 'call')))
   expect_equal(attr(f1, 'name'), 'cb.print.evaluation')
@@ -57,13 +61,13 @@ test_that("cb.print.evaluation works as expected", {
   expect_output(f1(), "\\[7\\]\ttrain-auc:0.900000\ttest-auc:0.800000")
   expect_output(f5(), "\\[7\\]\ttrain-auc:0.900000\ttest-auc:0.800000")
 
-  bst_evaluation_err  <- c('train-auc'=0.1, 'test-auc'=0.2)
+  bst_evaluation_err  <- c('train-auc' = 0.1, 'test-auc' = 0.2)
   expect_output(f1(), "\\[7\\]\ttrain-auc:0.900000\\+0.100000\ttest-auc:0.800000\\+0.200000")
 })
 
 test_that("cb.evaluation.log works as expected", {
 
-  bst_evaluation <- c('train-auc'=0.9, 'test-auc'=0.8)
+  bst_evaluation <- c('train-auc' = 0.9, 'test-auc' = 0.8)
   bst_evaluation_err <- NULL
 
   evaluation_log <- list()
@@ -75,37 +79,38 @@ test_that("cb.evaluation.log works as expected", {
   iteration <- 1
   expect_silent(f())
   expect_equal(evaluation_log,
-               list(c(iter=1, bst_evaluation)))
+               list(c(iter = 1, bst_evaluation)))
   iteration <- 2
   expect_silent(f())
   expect_equal(evaluation_log,
-               list(c(iter=1, bst_evaluation), c(iter=2, bst_evaluation)))
+               list(c(iter = 1, bst_evaluation), c(iter = 2, bst_evaluation)))
   expect_silent(f(finalize = TRUE))
   expect_equal(evaluation_log,
-               data.table(iter=1:2, train_auc=c(0.9,0.9), test_auc=c(0.8,0.8)))
+               data.table(iter = 1:2, train_auc = c(0.9, 0.9), test_auc = c(0.8, 0.8)))
 
-  bst_evaluation_err  <- c('train-auc'=0.1, 'test-auc'=0.2)
+  bst_evaluation_err  <- c('train-auc' = 0.1, 'test-auc' = 0.2)
   evaluation_log <- list()
   f <- cb.evaluation.log()
 
   iteration <- 1
   expect_silent(f())
   expect_equal(evaluation_log,
-               list(c(iter=1, c(bst_evaluation, bst_evaluation_err))))
+               list(c(iter = 1, c(bst_evaluation, bst_evaluation_err))))
   iteration <- 2
   expect_silent(f())
   expect_equal(evaluation_log,
-               list(c(iter=1, c(bst_evaluation, bst_evaluation_err)),
-                    c(iter=2, c(bst_evaluation, bst_evaluation_err))))
+               list(c(iter = 1, c(bst_evaluation, bst_evaluation_err)),
+                    c(iter = 2, c(bst_evaluation, bst_evaluation_err))))
   expect_silent(f(finalize = TRUE))
   expect_equal(evaluation_log,
-               data.table(iter=1:2,
-                          train_auc_mean=c(0.9,0.9), train_auc_std=c(0.1,0.1),
-                          test_auc_mean=c(0.8,0.8), test_auc_std=c(0.2,0.2)))
+               data.table(iter = 1:2,
+                          train_auc_mean = c(0.9, 0.9), train_auc_std = c(0.1, 0.1),
+                          test_auc_mean = c(0.8, 0.8), test_auc_std = c(0.2, 0.2)))
 })
 
 
-param <- list(objective = "binary:logistic", max_depth = 4, nthread = 2)
+param <- list(objective = "binary:logistic", eval_metric = "error",
+              max_depth = 4, nthread = n_threads)
 
 test_that("can store evaluation_log without printing", {
   expect_silent(
@@ -173,16 +178,18 @@ test_that("cb.reset.parameters works as expected", {
 })
 
 test_that("cb.save.model works as expected", {
-  files <- c('xgboost_01.model', 'xgboost_02.model', 'xgboost.model')
+  files <- c('xgboost_01.json', 'xgboost_02.json', 'xgboost.json')
   for (f in files) if (file.exists(f)) file.remove(f)
 
   bst <- xgb.train(param, dtrain, nrounds = 2, watchlist, eta = 1, verbose = 0,
-                   save_period = 1, save_name = "xgboost_%02d.model")
-  expect_true(file.exists('xgboost_01.model'))
-  expect_true(file.exists('xgboost_02.model'))
-  b1 <- xgb.load('xgboost_01.model')
+                   save_period = 1, save_name = "xgboost_%02d.json")
+  expect_true(file.exists('xgboost_01.json'))
+  expect_true(file.exists('xgboost_02.json'))
+  b1 <- xgb.load('xgboost_01.json')
+  xgb.parameters(b1) <- list(nthread = 2)
   expect_equal(xgb.ntree(b1), 1)
-  b2 <- xgb.load('xgboost_02.model')
+  b2 <- xgb.load('xgboost_02.json')
+  xgb.parameters(b2) <- list(nthread = 2)
   expect_equal(xgb.ntree(b2), 2)
 
   xgb.config(b2) <- xgb.config(bst)
@@ -191,9 +198,9 @@ test_that("cb.save.model works as expected", {
 
   # save_period = 0 saves the last iteration's model
   bst <- xgb.train(param, dtrain, nrounds = 2, watchlist, eta = 1, verbose = 0,
-                   save_period = 0)
-  expect_true(file.exists('xgboost.model'))
-  b2 <- xgb.load('xgboost.model')
+                   save_period = 0, save_name = 'xgboost.json')
+  expect_true(file.exists('xgboost.json'))
+  b2 <- xgb.load('xgboost.json')
   xgb.config(b2) <- xgb.config(bst)
   expect_equal(bst$raw, b2$raw)
 
@@ -236,8 +243,8 @@ test_that("early stopping xgb.train works", {
 test_that("early stopping using a specific metric works", {
   set.seed(11)
   expect_output(
-    bst <- xgb.train(param, dtrain, nrounds = 20, watchlist, eta = 0.6,
-                     eval_metric="logloss", eval_metric="auc",
+    bst <- xgb.train(param[-2], dtrain, nrounds = 20, watchlist, eta = 0.6,
+                     eval_metric = "logloss", eval_metric = "auc",
                      callbacks = list(cb.early.stop(stopping_rounds = 3, maximize = FALSE,
                                                     metric_name = 'test_logloss')))
   , "Stopping. Best iteration")
@@ -252,6 +259,27 @@ test_that("early stopping using a specific metric works", {
   expect_equal(logloss_log, logloss_pred, tolerance = 1e-5)
 })
 
+test_that("early stopping works with titanic", {
+  # This test was inspired by https://github.com/dmlc/xgboost/issues/5935
+  # It catches possible issues on noLD R
+  titanic <- titanic::titanic_train
+  titanic$Pclass <-  as.factor(titanic$Pclass)
+  dtx <- model.matrix(~ 0 + ., data = titanic[, c("Pclass", "Sex")])
+  dty <- titanic$Survived
+
+  xgboost::xgboost(
+    data = dtx,
+    label = dty,
+    objective = "binary:logistic",
+    eval_metric = "auc",
+    nrounds = 100,
+    early_stopping_rounds = 3,
+    nthread = n_threads
+  )
+
+  expect_true(TRUE)  # should not crash
+})
+
 test_that("early stopping xgb.cv works", {
   set.seed(11)
   expect_output(
@@ -267,12 +295,12 @@ test_that("early stopping xgb.cv works", {
 
 test_that("prediction in xgb.cv works", {
   set.seed(11)
-  nrounds = 4
+  nrounds <- 4
   cv <- xgb.cv(param, dtrain, nfold = 5, eta = 0.5, nrounds = nrounds, prediction = TRUE, verbose = 0)
   expect_false(is.null(cv$evaluation_log))
   expect_false(is.null(cv$pred))
   expect_length(cv$pred, nrow(train$data))
-  err_pred <- mean( sapply(cv$folds, function(f) mean(err(ltrain[f], cv$pred[f]))) )
+  err_pred <- mean(sapply(cv$folds, function(f) mean(err(ltrain[f], cv$pred[f]))))
   err_log <- cv$evaluation_log[nrounds, test_error_mean]
   expect_equal(err_pred, err_log, tolerance = 1e-6)
 
@@ -287,7 +315,7 @@ test_that("prediction in xgb.cv works", {
 
 test_that("prediction in xgb.cv works for gblinear too", {
   set.seed(11)
-  p <- list(booster = 'gblinear', objective = "reg:logistic", nthread = 2)
+  p <- list(booster = 'gblinear', objective = "reg:logistic", nthread = n_threads)
   cv <- xgb.cv(p, dtrain, nfold = 5, eta = 0.5, nrounds = 2, prediction = TRUE, verbose = 0)
   expect_false(is.null(cv$evaluation_log))
   expect_false(is.null(cv$pred))
@@ -308,7 +336,7 @@ test_that("prediction in early-stopping xgb.cv works", {
   expect_false(is.null(cv$pred))
   expect_length(cv$pred, nrow(train$data))
 
-  err_pred <- mean( sapply(cv$folds, function(f) mean(err(ltrain[f], cv$pred[f]))) )
+  err_pred <- mean(sapply(cv$folds, function(f) mean(err(ltrain[f], cv$pred[f]))))
   err_log <- cv$evaluation_log[cv$best_iteration, test_error_mean]
   expect_equal(err_pred, err_log, tolerance = 1e-6)
   err_log_last <- cv$evaluation_log[cv$niter, test_error_mean]
@@ -320,7 +348,7 @@ test_that("prediction in xgb.cv for softprob works", {
   set.seed(11)
   expect_warning(
     cv <- xgb.cv(data = as.matrix(iris[, -5]), label = lb, nfold = 4,
-                 eta = 0.5, nrounds = 5, max_depth = 3, nthread = 2,
+                 eta = 0.5, nrounds = 5, max_depth = 3, nthread = n_threads,
                  subsample = 0.8, gamma = 2, verbose = 0,
                  prediction = TRUE, objective = "multi:softprob", num_class = 3)
   , NA)
